from __future__ import annotations

import os
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

from fs_utils import normalize_path
from utils import HashState

Env = Dict[str, Union[str, List[str]]]


@dataclass
class ContextualRelativePath:
    path: str
    context: Context

    def resolve(self):
        return str(self)

    def __str__(self):
        return os.path.relpath(
            Path(self.context.repo_root).joinpath(self.path), self.context.cwd
        )


class Context(ABC):
    def __init__(self, repo_root: str, cwd: str):
        self.repo_root = repo_root
        self.cwd = os.path.abspath(cwd)

    def absolute(self, path: str):
        return normalize_path(self.repo_root, self.cwd, path)

    def relative(self, path: str):
        return ContextualRelativePath(self.absolute(path), self)

    def chdir(self, dest: str):
        self.cwd = os.path.abspath(
            Path(self.repo_root).joinpath(
                normalize_path(self.repo_root, self.cwd, dest)
            )
        )

    def sh(self, cmd: str, env: Env = None):
        raise NotImplementedError

    def add_dep(self, dep: str):
        self.add_deps([dep])

    def add_deps(self, deps: Sequence[str]):
        raise NotImplementedError

    def input(self, *, file: Optional[str], sh: Optional[str], env: Env = None):
        raise NotImplementedError


class MemorizeContext(Context):
    def __init__(self, repo_root: str, cwd: str, hashstate: HashState):
        super().__init__(repo_root, cwd)
        self.hashstate = hashstate
        self.inputs = []

    def chdir(self, dest: str):
        super().chdir(dest)
        self.hashstate.record("chdir", dest)

    def sh(self, cmd: str, env: Env = None):
        self.hashstate.record("sh", cmd, env)

    def add_deps(self, deps: Sequence[str]):
        self.hashstate.record("add_deps", deps)
        for dep in deps:
            if dep.startswith(":"):
                continue
            self.inputs.append(self.absolute(dep))

    def input(self, *, file: Optional[str], sh: Optional[str], env: Env = None):
        self.hashstate.record("input", file, sh, env)
        if file is not None:
            self.inputs.append(self.absolute(file))



import hashlib
import os
from pathlib import Path

from state import Rule
from utils import CacheMiss


def make_cache_fetcher(cache_directory: str):
    def cache_fetcher(state: str, target: str) -> str:
        dest = Path(cache_directory).joinpath(state).joinpath(target)
        try:
            with open(dest, "r") as f:
                return f.read()
        except FileNotFoundError:
            raise CacheMiss

    return cache_fetcher


def make_cache_memorize(cache_directory: str):
    def memorize(state: str, target: str, data: str):
        cache_target = Path(cache_directory).joinpath(state).joinpath(target)
        os.makedirs(os.path.dirname(cache_target), exist_ok=True)
        cache_target.write_text(data)

    return memorize


def get_cache_output_paths(cache_directory: str, rule: Rule, cache_key: str):
    cache_location = Path(cache_directory).joinpath(cache_key)
    keys = [
        hashlib.md5(output_path.encode("utf-8")).hexdigest()
        for output_path in rule.outputs
    ]
    for i, (output_path, key) in enumerate(zip(rule.outputs, keys)):
        if output_path.endswith("/"):
            keys[i] += "/"
    return cache_location, keys



import hashlib


class BuildException(Exception):
    pass


# thrown internally by contexts to exit early, should always be caught
class CacheMiss(Exception):
    pass


class MissingDependency(Exception):
    def __init__(self, *paths: str):
        self.paths = paths


class HashState:
    def __init__(self):
        self._state = hashlib.md5()

    def update(self, data: bytes):
        self._state.update(str(len(data)).encode("utf-8") + b":")
        self._state.update(data + b":")
        return self

    def record(self, *args):
        self._state.update(str(args).encode("utf-8"))
        return self

    def state(self):
        return self._state.hexdigest()



from typing import Callable, Union

from cache import make_cache_fetcher
from context import Env, MemorizeContext
from fs_utils import hash_file
from monitoring import log
from state import BuildState, Rule
from utils import CacheMiss, HashState, MissingDependency


class PreviewContext(MemorizeContext):
    def __init__(
        self,
        repo_root: str,
        cwd: str,
        hashstate: HashState,
        dep_fetcher: Callable[[str], str],
        cache_fetcher: Callable[[str, str], str],
    ):
        super().__init__(repo_root, cwd, hashstate)

        self.dep_fetcher = dep_fetcher
        self.cache_fetcher = cache_fetcher

    def input(self, *, file: str = None, sh: str = None, env: Env = None):
        state = self.hashstate.state()
        super().input(file=file, sh=sh, env=env)
        if file is not None:
            return self.dep_fetcher(self.absolute(file))
        else:
            return self.cache_fetcher(state, HashState().record(sh, env).state())


def make_dep_fetcher(build_state: BuildState):
    def dep_fetcher(input_path, *, get_hash=False) -> Union[str, bytes]:
        try:
            if input_path not in build_state.source_files:
                rule = build_state.target_rule_lookup.lookup(build_state, input_path)
                # this input may be stale / unbuilt
                # if so, do not read it, but instead throw MissingDependency
                if rule not in build_state.ready:
                    raise MissingDependency(input_path)
                # so it's already ready for use!

            if get_hash:
                return hash_file(input_path)
            else:
                with open(input_path) as f:
                    return f.read()
        except FileNotFoundError:
            raise MissingDependency(input_path)

    return dep_fetcher


def get_deps(build_state: BuildState, rule: Rule):
    """
    Use static dependencies and caches to try and identify as *many*
    needed dependencies as possible, without *any* spurious dependencies.
    """
    hashstate = HashState()
    cache_fetcher = make_cache_fetcher(build_state.cache_directory)
    dep_fetcher = make_dep_fetcher(build_state)

    log(f"Looking for static dependencies of {rule}")
    for dep in rule.deps:
        if dep not in build_state.source_files:
            if (
                build_state.target_rule_lookup.lookup(build_state, dep)
                not in build_state.ready
            ):
                log(f"Static dependency {dep} of {rule} is not ready, skipping impl")
                # static deps are not yet ready
                break
        if dep.startswith(":"):
            continue
        hashstate.update(dep.encode("utf-8"))
        try:
            hashstate.update(dep_fetcher(dep, get_hash=True))
        except MissingDependency:
            # get static deps before running the impl!
            # this means that a source file is *missing*, but the error will be thrown in enqueue_deps
            break
    else:
        ctx = PreviewContext(
            build_state.repo_root,
            rule.location,
            hashstate,
            dep_fetcher,
            cache_fetcher,
        )
        ok = False
        try:
            log(f"Running impl of {rule} to discover dynamic dependencies")
            rule.impl(ctx)
            log(f"Impl of {rule} completed with discovered deps: {ctx.inputs}")
            for out in rule.outputs:
                # needed so that if we ask for another output, we don't panic if it's not in the cache
                hashstate.record(out)
            ok = True
        except CacheMiss:
            log(f"Cache miss while running impl of {rule}")
            pass  # stops context execution
        except MissingDependency as e:
            log(f"Dependencies {e.paths} were unavailable while running impl of {rule}")
            pass  # dep already added to ctx.inputs
        # if `ok`, hash loaded dynamic dependencies
        if ok:
            log(
                f"Runtime dependencies resolved for {rule}, now checking dynamic dependencies"
            )
            for input_path in ctx.inputs:
                hashstate.update(input_path.encode("utf-8"))
                try:
                    data = dep_fetcher(input_path, get_hash=True)
                except MissingDependency as e:
                    # this dependency was not needed for deps calculation
                    # but is not verified to be up-to-date
                    ok = False
                    log(
                        f"Dynamic dependencies {e.paths} were not needed for the impl, but are not up to date"
                    )
                    break
                else:
                    hashstate.update(data)
        return (
            hashstate.state() if ok else None,
            ctx.inputs + rule.deps,
        )
    return None, rule.deps



import sys
import pathlib
from typing import Callable, Optional, Sequence, Union

sys.path.append(str(pathlib.Path(__file__).parent.absolute()))


"""
Below are some stubs to make imports in BUILD and rules.py files more pleasant
see loader.py for injected implementation
"""


def load(path: str):
    ...


def callback(
    *,
    name: Optional[str] = None,
    deps: Sequence[str] = (),
    impl: Callable,
    out: Union[str, Sequence[str]] = (),
):
    ...


def find(path: str):
    ...



import hashlib
import os
from pathlib import Path
from shutil import SameFileError, copyfile, copytree
from typing import List

from utils import BuildException

from common.shell_utils import sh


def find_root():
    repo_root = os.path.abspath(os.path.curdir)
    while True:
        if "WORKSPACE" in os.listdir(repo_root):
            return repo_root
        repo_root = os.path.dirname(repo_root)
        if repo_root == os.path.dirname(repo_root):
            break
    raise BuildException(
        "Unable to find WORKSPACE file - are you in the project directory?"
    )


def get_repo_files() -> List[str]:
    return [
        file.decode("ascii") if isinstance(file, bytes) else file
        for file in sh(
            "git", "ls-files", "--exclude-standard", capture_output=True, quiet=True
        ).splitlines()  # All tracked files
        + sh(
            "git",
            "ls-files",
            "-o",
            "--exclude-standard",
            capture_output=True,
            quiet=True,
        ).splitlines()  # Untracked but not ignored files
    ]


def normalize_path(repo_root, build_root, path):
    suffix = "/" if path.endswith("/") else ""
    if path.startswith("//"):
        path = Path(repo_root).joinpath(path[2:])
    else:
        path = Path(repo_root).joinpath(build_root, path)
    path = Path(os.path.abspath(path))
    repo_root = Path(os.path.abspath(repo_root))
    if repo_root not in path.parents:
        raise BuildException(
            f"Target `{path}` is not in the root directory of the repo."
        )
    return str(path.relative_to(repo_root)) + suffix


def copy_helper(*, src_root, dest_root, src_names, dest_names=None, symlink=False):
    if not dest_names:
        dest_names = src_names
    assert len(src_names) == len(dest_names)
    for src_name, dest_name in zip(src_names, dest_names):
        src = Path(src_root).joinpath(src_name)
        dest = Path(dest_root).joinpath(dest_name)
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        try:
            if symlink:
                Path(dest).symlink_to(src)
            elif src_name.endswith("/"):
                copytree(src=src, dst=dest, dirs_exist_ok=True)
            else:
                copyfile(src, dest)
        except SameFileError:
            pass


def hash_file(path):
    # only hash files whose contents are "locked in". That way we can cache safely.
    if path in hash_file.cache:
        return hash_file.cache[path]
    else:
        with open(path, "rb") as f:
            out = hash_file.cache[path] = (
                hashlib.md5(f.read()).hexdigest().encode("utf-8")
            )
            return out


hash_file.cache = {}



from __future__ import annotations

from collections import Collection

from state import BuildState, Rule
from monitoring import log


def enqueue_deps(
    build_state: BuildState, rule: Rule, candidate_deps: Collection[str]
) -> bool:
    waiting_for_deps = False

    with build_state.scheduling_lock:
        for dep in candidate_deps:
            if dep in build_state.source_files:
                # nothing to do
                continue

            runtime_dep: Rule = build_state.target_rule_lookup.lookup(build_state, dep)

            if runtime_dep not in build_state.ready:
                waiting_for_deps = True
                if runtime_dep not in build_state.scheduled_but_not_ready:
                    # enqueue dependency
                    log(f"Enqueueing dependency {runtime_dep}")
                    build_state.scheduled_but_not_ready.add(runtime_dep)
                    build_state.work_queue.put(runtime_dep)
                    build_state.status_monitor.move(total=1)
                else:
                    log(f"Waiting on already queued dependency {runtime_dep}")
                # register task in the already queued / executing dependency
                # so when it finishes we may be triggered
                runtime_dep.runtime_dependents.append(rule)
                rule.pending_rule_dependencies.append(runtime_dep)
    return waiting_for_deps



from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()


setup(
    name="buildtool",
    version="0.0.0b175.dev21",
    author="Rahul Arya",
    author_email="rahularya@berkeley.edu",
    long_description=readme,
    long_description_content_type="text/markdown",
    licence="MIT",
    packages=find_packages(
        include=["buildtool", "buildtool.common", "buildtool.common.rpc"]
    ),
    package_data={"": ["**/*.tex"]},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "buildtool=buildtool.__main__:cli",
            "bt=buildtool.__main__:cli",
        ]
    },
    python_requires=">=3.8",
    install_requires=["click", "flask", "cachetools", "colorama", "requests", "tqdm"],
)
